"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var parseImmediate_1 = require("../utils/parseImmediate");
exports.directives = [
    '.word',
    '.asciiz',
    '.ascii',
    '.byte',
    '.space',
    '.float',
    '.double',
];
function fromAST(node) {
    var directive = node.dataDirective;
    if (!~exports.directives.indexOf(directive)) {
        throw new Error("Unknown data directive: " + directive);
    }
    switch (directive) {
        case '.ascii':
        case '.asciiz':
            var str = node.values[0].replace(/^"|"$/g, '');
            return directive === '.ascii' ? new ASCII(str) : new ASCIIZ(str);
        case '.word':
            return new Word(node.values.map(parseImmediate_1.default));
        case '.byte':
            return new Byte(node.values.map(function (value) {
                if (/^'/.test(value)) {
                    return value.charCodeAt(1);
                }
                return parseImmediate_1.default(value);
            }));
        case '.space':
            return new Byte(new Array(parseImmediate_1.default(node.values[0])).fill(0));
    }
    // TODO: Implement
    return new Word([1]);
}
exports.fromAST = fromAST;
var Byte = /** @class */ (function () {
    function Byte(values) {
        this.type = 'byte';
        this.values = values;
    }
    Byte.prototype.size = function () {
        return this.values.length;
    };
    Byte.prototype.toInt8Array = function () {
        var arr = new Uint8Array(this.size());
        for (var i = 0; i < this.values.length; i++) {
            arr[i] = this.values[i] | 0;
        }
        return arr;
    };
    return Byte;
}());
exports.Byte = Byte;
var Word = /** @class */ (function () {
    function Word(values) {
        this.type = 'word';
        this.values = values;
    }
    Word.prototype.size = function () {
        return this.values.length * 4;
    };
    Word.prototype.toInt8Array = function () {
        var arr = new Uint8Array(this.size());
        for (var i = 0; i < this.values.length; i++) {
            arr[i + 0] = (this.values[i] & 0xff000000) >>> 24;
            arr[i + 1] = (this.values[i] & 0x00ff0000) >>> 16;
            arr[i + 2] = (this.values[i] & 0x0000ff00) >>> 8;
            arr[i + 3] = (this.values[i] & 0x000000ff);
        }
        return arr;
    };
    return Word;
}());
exports.Word = Word;
var BaseASCII = /** @class */ (function () {
    function BaseASCII(value, includeNullTerminator) {
        if (includeNullTerminator === void 0) { includeNullTerminator = false; }
        this.includeNullTerminator = false;
        this.value = value;
        this.includeNullTerminator = includeNullTerminator;
    }
    BaseASCII.prototype.size = function () {
        return this.value.length + (this.includeNullTerminator ? 1 : 0);
    };
    // TODO: Throw error when UTF8/Unicode chars are encountered?
    BaseASCII.prototype.toInt8Array = function () {
        var size = this.size();
        var arr = new Uint8Array(size);
        for (var i = 0; i < this.value.length; i++) {
            arr[i] = this.value.charCodeAt(i);
        }
        if (this.includeNullTerminator) {
            arr[size - 1] = 0;
        }
        return arr;
    };
    return BaseASCII;
}());
exports.BaseASCII = BaseASCII;
var ASCII = /** @class */ (function (_super) {
    tslib_1.__extends(ASCII, _super);
    function ASCII(value) {
        var _this = _super.call(this, value, false) || this;
        _this.type = 'ascii';
        return _this;
    }
    return ASCII;
}(BaseASCII));
exports.ASCII = ASCII;
var ASCIIZ = /** @class */ (function (_super) {
    tslib_1.__extends(ASCIIZ, _super);
    function ASCIIZ(value) {
        var _this = _super.call(this, value, true) || this;
        _this.type = 'asciiz';
        return _this;
    }
    return ASCIIZ;
}(BaseASCII));
exports.ASCIIZ = ASCIIZ;
