"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var AST = require("../parse/AST");
var Instruction_1 = require("../Instruction");
var mappings_1 = require("../utils/mappings");
var parseImmediate_1 = require("../utils/parseImmediate");
var Data_1 = require("../Data");
var AssmebledObject_1 = require("./AssmebledObject");
var BASE_TEXT_ADDR = 0x00400000;
var BASE_DATA_ADDR = 0x10010000;
var invertIndexTable = function (t) {
    var ret = {};
    Object.keys(t).forEach(function (k) {
        ret[t[k]] = k;
    });
    return ret;
};
var Assembler = /** @class */ (function () {
    function Assembler(root) {
        this.instructionLabels = {};
        this.dataLabels = {};
        this.symbols = {};
        this.relocations = {};
        this.astRoot = root;
    }
    Assembler.prototype.genData = function () {
        var _this = this;
        var dataIndex = 0;
        var allData = this.astRoot.data.children.map(function (node) {
            if (AST.utils.isDataNode(node)) {
                var data = Data_1.fromAST(node);
                dataIndex += data.size();
                return data;
            }
            else if (AST.utils.isLabelNode(node)) {
                var label = node.name;
                var firstChild = node.children[0];
                if (firstChild && AST.utils.isDataNode(firstChild)) {
                    var data = Data_1.fromAST(firstChild);
                    _this.dataLabels[label] = dataIndex;
                    dataIndex += data.size();
                    return data;
                }
                else {
                    throw new Error('This should not happen');
                }
            }
            else {
                throw new Error('This should not happen');
            }
        });
        var size = allData.reduce(function (a, b) { return a + b.size(); }, 0);
        this.data = new Uint8Array(size);
        var index = 0;
        allData.forEach(function (data) {
            var segment = data.toInt8Array();
            for (var i = 0; i < segment.length; i++) {
                _this.data[index++] = segment[i];
            }
        });
    };
    Assembler.prototype.genASTInstructions = function () {
        var _this = this;
        this.astInstructions = [];
        var index = 0;
        var walk = function (child) {
            if (AST.utils.isOperationNode(child)) {
                _this.astInstructions[index++] = child;
            }
            else if (AST.utils.isLabelNode(child)) {
                _this.instructionLabels[child.name] = index;
                child.children.forEach(walk);
            }
            else if (AST.utils.isTranformedNode(child)) {
                child.children.forEach(walk);
            }
            else {
                throw new Error('This should not happen');
            }
        };
        this.astRoot.text.children.forEach(walk);
    };
    Assembler.prototype.indexOfLabel = function (node) {
        for (var i = 0; i < node.args.length; i++) {
            if (AST.utils.isAddressNode(node.args[i]))
                return i;
        }
        return -1;
    };
    Assembler.prototype.genSymbolTable = function () {
        var _this = this;
        var invertedInstructionLabels = invertIndexTable(this.instructionLabels);
        this.symbols = {};
        this.astInstructions.forEach(function (op, i) {
            if (i in invertedInstructionLabels) {
                var addr = BASE_TEXT_ADDR + i * 4;
                _this.symbols[invertedInstructionLabels[i]] = addr;
            }
        });
        Object.keys(this.dataLabels).forEach(function (label) {
            _this.symbols[label] = BASE_DATA_ADDR + _this.dataLabels[label];
        });
    };
    Assembler.prototype.genRelocationTable = function () {
        var _this = this;
        this.relocations = {};
        this.astInstructions.forEach(function (op, i) {
            var index = _this.indexOfLabel(op);
            if ((op.name === 'j' || op.name == 'jal') && ~index) {
                var addrNode = op.args[index];
                if ((addrNode.label in _this.symbols)) {
                    var addr = BASE_TEXT_ADDR + i * 4;
                    _this.relocations[addr] = addrNode.label;
                }
                else {
                    // TODO: better error
                    throw new Error("Invalid Jump");
                }
            }
        });
    };
    Assembler.prototype.interpolateLabels = function () {
        var _this = this;
        this.astInstructions.forEach(function (node, i) {
            var index = _this.indexOfLabel(node);
            if (~index) {
                var addrNode = node.args[index];
                var label = addrNode.label;
                if (!(label in _this.symbols)) {
                    throw new Error("Missing Symbol: \"" + label + "\"");
                }
                var addr = _this.symbols[label];
                var immediateValue = 0;
                if (node.name === 'beq' || node.name === 'bne') {
                    // TODO: Validiate?
                    //  1. It's an address of an instruction
                    //  2. It's not a partial
                    // pc address
                    var pcAddress = BASE_TEXT_ADDR + (i + 1) * 4;
                    var offset = addr - pcAddress;
                    immediateValue = offset / 4;
                }
                else {
                    if (addrNode.partitioned) {
                        immediateValue = addrNode.part === 'upper'
                            ? (addr & 0xffff0000) >>> 16
                            : (addr & 0x0000ffff);
                    }
                    else {
                        immediateValue = addr;
                    }
                }
                node.args[index] = new AST.ImmediateNode(immediateValue.toString());
            }
        });
    };
    Assembler.prototype.genObjInstructions = function () {
        this.objInstructions = this.astInstructions.map(function (node) {
            var info = mappings_1.instructions[node.name];
            // TODO: Better errors
            if (!info)
                throw new Error("Assmebler: Unknown instruction \"" + node.name + "\"");
            var rs;
            var rt;
            var rd;
            if (info.type === 'R') {
                switch (node.name) {
                    case 'syscall':
                        return new Instruction_1.RInstruction(0, 0, 0, 0, 0, info.code);
                    case 'jr':
                    case 'jalr':
                        rs = mappings_1.registers[node.args[0].name];
                        return new Instruction_1.RInstruction(0, rs, 0, 0, 0, info.code);
                    case 'mfhi':
                    case 'mflo':
                        rd = mappings_1.registers[node.args[0].name];
                        return new Instruction_1.RInstruction(0, 0, 0, rd, 0, info.code);
                    case 'mult':
                    case 'div':
                    case 'multu':
                    case 'divu':
                        rs = mappings_1.registers[node.args[0].name];
                        rt = mappings_1.registers[node.args[1].name];
                        return new Instruction_1.RInstruction(0, rs, rt, 0, 0, info.code);
                    case 'sll':
                    case 'srl':
                    case 'sra':
                        rd = mappings_1.registers[node.args[0].name];
                        rt = mappings_1.registers[node.args[1].name];
                        var sh = parseImmediate_1.default(node.args[2].value);
                        return new Instruction_1.RInstruction(0, 0, rt, 0, sh, info.code);
                    case 'sllv':
                    case 'srlv':
                    case 'srav':
                        rd = mappings_1.registers[node.args[0].name];
                        rt = mappings_1.registers[node.args[1].name];
                        rs = mappings_1.registers[node.args[2].name];
                        return new Instruction_1.RInstruction(0, rs, rt, rd, 0, info.code);
                    default:
                        rd = mappings_1.registers[node.args[0].name];
                        rs = mappings_1.registers[node.args[1].name];
                        rt = mappings_1.registers[node.args[2].name];
                        return new Instruction_1.RInstruction(0, rs, rt, rd, 0, info.code);
                }
            }
            if (info.type === 'I') {
                var imm = void 0;
                switch (node.name) {
                    case 'lui':
                        rt = mappings_1.registers[node.args[0].name];
                        imm = parseImmediate_1.default(node.args[1].value);
                        return new Instruction_1.IInstruction(info.code, 0, rt, imm);
                    case 'lb':
                    case 'lbu':
                    case 'lh':
                    case 'lhu':
                    case 'lw':
                    case 'sb':
                    case 'sh':
                    case 'sw':
                        rt = mappings_1.registers[node.args[0].name];
                        imm = parseImmediate_1.default(node.args[1].offset);
                        rs = mappings_1.registers[node.args[1].register];
                        return new Instruction_1.IInstruction(info.code, rs, rt, imm);
                    default:
                        rt = mappings_1.registers[node.args[0].name];
                        rs = mappings_1.registers[node.args[1].name];
                        imm = parseImmediate_1.default(node.args[2].value);
                        return new Instruction_1.IInstruction(info.code, rs, rt, imm);
                }
            }
            if (info.type === 'J') {
                var addr = parseImmediate_1.default(node.args[0].value);
                return new Instruction_1.JInstruction(info.code, addr);
            }
            // TODO: Better errors
            throw new Error('This should not happen');
        });
    };
    Assembler.prototype.genInstructions = function () {
        var _this = this;
        this.instructions = new Uint32Array(this.objInstructions.length);
        this.objInstructions.forEach(function (obj, i) {
            _this.instructions[i] = obj.toInt();
        });
    };
    Assembler.prototype.assemble = function () {
        var ret = new AssmebledObject_1.default();
        this.genData();
        this.genASTInstructions();
        this.genSymbolTable();
        this.genRelocationTable();
        this.interpolateLabels();
        this.genObjInstructions();
        this.genInstructions();
        ret.symbols = this.symbols;
        ret.relocations = this.relocations;
        ret.data = this.data;
        ret.instructions = this.instructions;
        ret.objInstructions = this.objInstructions;
        return ret;
    };
    return Assembler;
}());
exports.default = Assembler;
