"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Token_1 = require("./Token");
var Lexer = /** @class */ (function () {
    function Lexer(source) {
        this.source = source;
        this.reset();
    }
    Lexer.prototype.reset = function () {
        this.lineno = 1;
        this.colno = 1;
        this.inComment = false;
        this.value = '';
        this.tokens = [];
        this.inQuote = false;
    };
    Lexer.prototype.normalizedTokens = function () {
        var ret = [];
        this.tokens.forEach(function (t, i) {
            if (t.type !== 'EOL' || (ret.length && ret[ret.length - 1].type !== 'EOL')) {
                if (t.type === 'EOS' && ret.length && ret[ret.length - 1].type === 'EOL') {
                    ret[ret.length - 1] = t;
                }
                else {
                    ret.push(t);
                }
            }
        });
        return ret;
    };
    Lexer.prototype.pushSpecial = function (type, value) {
        if (this.inComment)
            return;
        this.tokens.push(new Token_1.Token(type, value, this.lineno, this.colno));
    };
    Lexer.prototype.flush = function () {
        if (this.inComment)
            return;
        var value = this.value;
        if (!value)
            return;
        var token = new Token_1.Token('TEXT', value, this.lineno, this.colno - this.value.length);
        this.tokens.push(token);
        this.value = '';
    };
    Lexer.prototype.lex = function () {
        var _this = this;
        this.reset();
        this.source.split('').forEach(function (c) {
            if (!_this.inComment && (c === '\'' || c === '"')) {
                _this.value += c;
                if (_this.inQuote) {
                    var type = /^\'/.test(_this.value) ? 'CHAR' : 'STRING';
                    _this.tokens.push(new Token_1.Token(type, _this.value, _this.lineno, _this.colno));
                    _this.value = '';
                    _this.inQuote = false;
                }
                else {
                    _this.inQuote = true;
                }
                return;
            }
            if (_this.inQuote) {
                _this.value += c;
                return;
            }
            switch (c) {
                // Special-case tokens
                case ':':
                    _this.flush();
                    _this.pushSpecial('COLON', c);
                    break;
                case '(':
                    _this.flush();
                    _this.pushSpecial('LEFT_PAREN', c);
                    break;
                case ')':
                    _this.flush();
                    _this.pushSpecial('RIGHT_PAREN', c);
                    break;
                case ',':
                    _this.flush();
                    _this.pushSpecial('COMMA', c);
                    break;
                // Comment, whitespace, new-lines
                case '#':
                    _this.flush();
                    _this.inComment = true;
                    break;
                case ' ':
                case '\t':
                    _this.flush();
                    break;
                case '\n':
                    _this.flush();
                    if (_this.inComment) {
                        _this.inComment = false;
                    }
                    _this.pushSpecial('EOL', c);
                    _this.lineno++;
                    _this.colno = 0;
                    break;
                // Everything else
                default:
                    if (!_this.inComment) {
                        _this.value += c;
                    }
            }
            _this.colno++;
        });
        if (!this.inComment) {
            this.flush();
        }
        else {
            this.inComment = false;
        }
        this.pushSpecial('EOS', '%EOS%');
        return this.normalizedTokens();
    };
    return Lexer;
}());
exports.default = Lexer;
