"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var AST = require("./AST");
var transform_1 = require("./transform");
var rootLevelDirectives = [
    '.data',
    '.text',
];
var dataLevelDirectives = [
    '.word',
    '.asciiz',
    '.ascii',
    '.word',
    '.byte',
    '.space',
    '.float',
    '.double',
];
// Matches (leading - sign, or not) for hex, octal, binary, and decimal
var IMM_RE = /^(-?)(0x[0-9A-Fa-f]+|0o[0-7]+|0b[0-1]+|[0-9]+)$/;
// TODO: Make errors usefull, i.e. wrap in a class that includes lineno, colno
var Parser = /** @class */ (function () {
    function Parser(tokens) {
        this.tokens = tokens;
        this.index = 0;
        this.root = new AST.Root();
        this.inData = false;
    }
    Parser.prototype.peek = function (n) {
        if (n === void 0) { n = 0; }
        return this.tokens[this.index + n];
    };
    Parser.prototype.next = function () {
        return this.tokens[this.index++];
    };
    Parser.prototype.nextLine = function () {
        var tokens = [];
        var token = this.next();
        while (token && token.type !== 'EOS' && token.type !== 'EOL') {
            tokens.push(token);
            token = this.next();
        }
        return tokens;
    };
    Parser.prototype.addNode = function (node) {
        if (this.inData) {
            this.root.data.add(node);
        }
        else {
            this.root.text.add(node);
        }
    };
    Parser.prototype.groupByComma = function (tokens) {
        var ret = [];
        var buf = [];
        tokens.forEach(function (t, i) {
            if (t.type === 'COMMA') {
                if (buf.length) {
                    ret.push(buf);
                    buf = [];
                }
            }
            else {
                buf.push(t);
                if (i === tokens.length - 1) {
                    ret.push(buf);
                }
            }
        });
        return ret;
    };
    Parser.prototype.normalizeDataValues = function (tokens) {
        var ret = [];
        var group = this.groupByComma(tokens);
        return group.map(function (g) {
            if (g.length > 1) {
                throw new Error('TODO');
            }
            return g[0].value;
        });
    };
    Parser.prototype.normalizeArgs = function (tokens) {
        var group = this.groupByComma(tokens);
        return group.map(function (g) {
            var ret;
            if (g[0].type !== 'TEXT') {
                throw new Error('Unexpected token TODO');
            }
            if (/^\$/.test(g[0].value)) {
                // Register
                // TODO: Validate group
                ret = new AST.RegisterNode(g[0].value);
            }
            else if (IMM_RE.test(g[0].value)) {
                // Immediate OR Offset
                if (g.length === 1) {
                    // Immediate
                    ret = new AST.ImmediateNode(g[0].value);
                }
                else {
                    // Offset
                    var imm = g[0], lp = g[1], r = g[2], rp = g[3];
                    // TODO: Validate stuff
                    ret = new AST.OffsetNode(imm.value, r.value);
                }
            }
            else {
                // Address
                ret = new AST.AddressNode(g[0].value);
            }
            return ret;
        });
    };
    Parser.prototype.logTokens = function (tokens) {
        console.log(tokens.map(function (t) { return "(" + t.type + " => " + t.value + ")"; }).join(' '));
    };
    Parser.prototype.logGroup = function (group) {
        var _this = this;
        console.log('GROUP: (START)');
        group.forEach(function (tokens) { return _this.logTokens(tokens); });
        console.log('GROUP: (END)');
    };
    Parser.prototype.parse = function () {
        var line = this.nextLine();
        while (line.length) {
            var first = line[0], rest = line.slice(1);
            if (first.type !== 'TEXT') {
                throw new Error('Unexpected Token Error');
            }
            if (/^\./.test(first.value)) {
                // Directive
                if (!~rootLevelDirectives.indexOf(first.value)) {
                    throw new Error('Unknown directive');
                }
                if (rest.length) {
                    throw new Error('Unexpected tokens');
                }
                this.inData = first.value === '.data';
            }
            else {
                var node = void 0;
                var label = '';
                var labelNode = void 0;
                if (rest.length && rest[0].type === 'COLON') {
                    // Label
                    label = first.value;
                    labelNode = new AST.LabelNode(label);
                    if (rest.length > 1) {
                        first = rest[1];
                        rest = rest.slice(2);
                    }
                    else {
                        var nextLine = this.nextLine();
                        if (!nextLine.length) {
                            this.addNode(labelNode);
                            continue;
                        }
                        first = nextLine[0];
                        rest = nextLine.slice(1);
                    }
                }
                if (this.inData) {
                    if (first.type !== 'TEXT' && !/^\./.test(first.value)) {
                        throw new Error('Bad Data');
                    }
                    var dataDirective = first.value;
                    var values = this.normalizeDataValues(rest);
                    node = new AST.DataNode(dataDirective, values);
                }
                else {
                    var opName = first.value;
                    var args = this.normalizeArgs(rest);
                    node = new AST.OperationNode(first.value, args);
                }
                if (labelNode) {
                    labelNode.add(node);
                    this.addNode(labelNode);
                }
                else {
                    this.addNode(node);
                }
            }
            line = this.nextLine();
        }
        transform_1.default(this.root);
        return this.root;
    };
    return Parser;
}());
exports.default = Parser;
