"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var AST = require("./AST");
var $AT = new AST.RegisterNode('$at');
var $0 = new AST.RegisterNode('$0');
var transformImmediate = function (op, rOp) {
    var immNode = op.args[2];
    var split = immNode.split();
    // Immediate fits in 16 bits, bail
    if (split.length === 1)
        return op;
    // Pack the immediate value into the $at register, then invoke the associated R-type operation
    // with $at in replacement of immediate
    return new AST.TransformedNode([
        new AST.OperationNode('lui', [$AT, split[0]]),
        new AST.OperationNode('ori', [$AT, $AT, split[1]]),
        new AST.OperationNode(rOp, [op.args[0], op.args[1], $AT]),
    ], op);
};
var transformers = {
    // bge $rx, $ry, addr =>
    //    slt $at, $rx, $ry
    //    beq $at, $0, addr
    'bge': function (op) { return new AST.TransformedNode([
        new AST.OperationNode('slt', [$AT, op.args[0], op.args[1]]),
        new AST.OperationNode('beq', [$AT, $0, op.args[2]]),
    ], op); },
    // bgt $rx, $ry, addr =>
    //    slt $at, $ry, $rx
    //    bne $at, $0, addr
    'bgt': function (op) { return new AST.TransformedNode([
        new AST.OperationNode('slt', [$AT, op.args[1], op.args[0]]),
        new AST.OperationNode('bne', [$AT, $0, op.args[2]]),
    ], op); },
    // ble $rx, $ry, addr =>
    //    slt $at, $ry, $rx
    //    beq $at, $0, addr
    'ble': function (op) { return new AST.TransformedNode([
        new AST.OperationNode('slt', [$AT, op.args[1], op.args[0]]),
        new AST.OperationNode('beq', [$AT, $0, op.args[2]]),
    ], op); },
    // blt $rx, $ry, addr =>
    //    slt $at, $rx, $ry
    //    bne $at, $0, addr
    'blt': function (op) { return new AST.TransformedNode([
        new AST.OperationNode('slt', [$AT, op.args[0], op.args[1]]),
        new AST.OperationNode('bne', [$AT, $0, op.args[2]]),
    ], op); },
    // la $rx, addr =>
    //    lui $at, upper(addr)
    //    ori $rx, $at, lower(addr)
    'la': function (op) {
        var addr = op.args[1];
        return new AST.TransformedNode([
            new AST.OperationNode('lui', [$AT, addr.partition('upper')]),
            new AST.OperationNode('ori', [op.args[0], $AT, addr.partition('lower')]),
        ], op);
    },
    // move $rx, $ry =>
    //    addu $rx, $0, $ry
    'move': function (op) { return new AST.TransformedNode([
        new AST.OperationNode('addu', [op.args[0], $0, op.args[1]]),
    ], op); },
    // li $rx, imm =>
    //    addiu $rx, $0, imm
    // OR
    // li $rx, imm =>
    //    lui $at, upper(imm)
    //    ori $rx, $at, lower(imm)
    'li': function (op) {
        var immNode = op.args[1];
        var split = immNode.split();
        if (split.length == 1) {
            return new AST.TransformedNode([
                new AST.OperationNode('addiu', [op.args[0], $0, op.args[1]]),
            ], op);
        }
        return new AST.TransformedNode([
            new AST.OperationNode('lui', [$AT, split[0]]),
            new AST.OperationNode('ori', [op.args[0], $AT, split[1]]),
        ], op);
    },
    // addi $rt, $rs, imm ?=>
    //    lui $at, upper(imm)
    //    ori $at, $at, lower(imm)
    //    add $rt, $rs, $at
    'addi': function (op) { return transformImmediate(op, 'add'); },
    // addiu $rt, $rs, imm ?=>
    //    lui $at, upper(imm)
    //    ori $at, $at, lower(imm)
    //    addu $rt, $rs, $at
    'addiu': function (op) { return transformImmediate(op, 'addu'); },
    // andi $rt, $rs, imm ?=>
    //    lui $at, upper(imm)
    //    ori $at, $at, lower(imm)
    //    and $rt, $rs, $at
    'andi': function (op) { return transformImmediate(op, 'and'); },
    // ori $rt, $rs, imm ?=>
    //    lui $at, upper(imm)
    //    ori $at, $at, lower(imm)
    //    or $rt, $rs, $at
    'ori': function (op) { return transformImmediate(op, 'or'); },
    // xori $rt, $rs, imm ?=>
    //    lui $at, upper(imm)
    //    ori $at, $at, lower(imm)
    //    xor $rt, $rs, $at
    'xori': function (op) { return transformImmediate(op, 'xor'); },
    // slti $rt, $rs, imm ?=>
    //    lui $at, upper(imm)
    //    ori $at, $at, lower(imm)
    //    slt $rt, $rs, $at
    'slti': function (op) { return transformImmediate(op, 'slt'); },
    // slti $rt, $rs, imm ?=>
    //    lui $at, upper(imm)
    //    ori $at, $at, lower(imm)
    //    sltu $rt, $rs, $at
    'sltiu': function (op) { return transformImmediate(op, 'sltu'); },
    // sgt $rd, $rs, $rt => slt $rd, $rt, $rs
    'sgt': function (op) { return new AST.TransformedNode([
        new AST.OperationNode('slt', [op.args[0], op.args[2], op.args[1]]),
    ], op); },
    // seq $rd, $rs, $rt =>
    //    subu $rd, $rs, $rt
    //    ori  $at, $0, 1
    //    sltu $rd, $rd, $at
    'seq': function (op) { return new AST.TransformedNode([
        new AST.OperationNode('subu', op.args),
        new AST.OperationNode('ori', [$AT, $0, new AST.ImmediateNode('1')]),
        new AST.OperationNode('sltu', [op.args[0], op.args[0], $AT]),
    ], op); },
};
var isOperationNode = function (n) { return (n.type === 'OPERATION'); };
function transform(root) {
    root.text.children.forEach(function (child, i) {
        var node;
        var parent;
        var index;
        if (child.type === 'LABEL') {
            parent = child;
            node = child.children[0];
            index = 0;
        }
        else {
            node = child;
            parent = root.text;
            index = i;
        }
        if (isOperationNode(node)) {
            if (node.name in transformers) {
                var transformer = transformers[node.name];
                parent.children[index] = transformer(node);
            }
        }
    });
}
exports.default = transform;
